import logging
import psutil
import dask
import numpy as np
import argparse
import traceback
import copy
import time
import sys
import os
from casatasks import casalog
from casatools import msmetadata, table
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *
from meersolar.meerpipeline.flagging import single_ms_flag
from meersolar.meerpipeline.import_model import import_fluxcal_models

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def run_delaycal(
    msname="",
    field="",
    scan="",
    uvrange="",
    refant="",
    refantmode="flex",
    solint="inf",
    combine="",
    gaintable=[],
    gainfield=[],
    interp=[],
    n_threads=-1,
    dry_run=False,
):
    """
    Perform delay calibration
    """
    limit_threads(n_threads=n_threads)
    from meersolar.utils.calibration import delaycal
    from casatasks import gaincal

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    print(f"Performing delay calibration on : {msname}")
    caltable_prefix = os.path.basename(msname).split(".ms")[0]
    with suppress_casa_output():
        if uvrange == "":
            gaincal(
                vis=msname,
                caltable=caltable_prefix + ".kcal",
                field=str(field),
                scan=str(scan),
                uvrange="",
                refant=refant,
                refantmode=refantmode,
                solint=solint,
                gaintype="K",
                combine=combine,
                gaintable=gaintable,
                gainfield=gainfield,
                interp=interp,
            )
        else:
            delaycal(
                vis=msname,
                caltable=caltable_prefix + ".kcal",
                field=str(field),
                scan=str(scan),
                uvrange=uvrange,
                refant=refant,
                refantmode=refantmode,
                solint=solint,
                combine=combine,
                gaintable=gaintable,
                gainfield=gainfield,
                interp=interp,
            )
    return caltable_prefix + ".kcal"


def run_bandpass(
    msname="",
    field="",
    scan="",
    uvrange="",
    refant="",
    solint="inf",
    solnorm=False,
    combine="",
    gaintable=[],
    gainfield=[],
    interp=[],
    n_threads=-1,
    dry_run=False,
):
    """
    Perform bandpass calibration
    """
    limit_threads(n_threads=n_threads)
    from casatasks import bandpass, flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    print(f"Performing bandpass calibration on : {msname}")
    caltable_prefix = os.path.basename(msname).split(".ms")[0]
    with suppress_casa_output():
        bandpass(
            vis=msname,
            caltable=caltable_prefix + ".bcal",
            field=str(field),
            scan=str(scan),
            uvrange=uvrange,
            refant=refant,
            solint=solint,
            solnorm=solnorm,
            combine=combine,
            gaintable=gaintable,
            gainfield=gainfield,
            interp=interp,
        )
        flagdata(
            vis=caltable_prefix + ".bcal",
            mode="rflag",
            datacolumn="CPARAM",
            flagbackup=False,
        )
    return caltable_prefix + ".bcal"


def run_gaincal(
    msname="",
    field="",
    scan="",
    uvrange="",
    refant="",
    gaintype="G",
    solint="inf",
    calmode="ap",
    refantmode="flex",
    solmode="",
    smodel=[],
    rmsthresh=[],
    combine="",
    append=False,
    gaintable=[],
    gainfield=[],
    interp=[],
    n_threads=-1,
    dry_run=False,
):
    """
    Perform gain calibration
    """
    limit_threads(n_threads=n_threads)
    from casatasks import gaincal, flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    print(f"Performing gain calibration on : {msname}")
    caltable_prefix = os.path.basename(msname).split(".ms")[0]
    with suppress_casa_output():
        gaincal(
            vis=msname,
            caltable=caltable_prefix + ".gcal",
            field=str(field),
            scan=str(scan),
            uvrange=uvrange,
            refant=refant,
            refantmode=refantmode,
            solint=solint,
            combine=combine,
            gaintype=gaintype,
            calmode=calmode,
            solmode=solmode,
            rmsthresh=rmsthresh,
            smodel=smodel,
            append=append,
            gaintable=gaintable,
            gainfield=gainfield,
            interp=interp,
        )
    return caltable_prefix + ".gcal"


def run_leakagecal(
    msname="",
    field="",
    scan="",
    uvrange="",
    refant="",
    combine="",
    gaintable=[],
    gainfield=[],
    interp=[],
    n_threads=-1,
    dry_run=False,
):
    """
    Perform relative leakage calibration (pol-conversion calibration)
    """
    limit_threads(n_threads=n_threads)
    from casatasks import polcal, flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    print(f"Performing relative leakage calibration on : {msname}")
    caltable_prefix = os.path.basename(msname).split(".ms")[0]
    with suppress_casa_output():
        polcal(
            vis=msname,
            caltable=caltable_prefix + ".dcal",
            field=str(field),
            scan=str(scan),
            uvrange=uvrange,
            refant=refant,
            solint="inf,10MHz",
            combine=combine,
            poltype="Df",
            gaintable=gaintable,
            gainfield=gainfield,
            interp=interp,
        )
        flagdata(
            vis=caltable_prefix + ".dcal",
            mode="rflag",
            datacolumn="CPARAM",
            flagbackup=False,
        )
    return caltable_prefix + ".dcal"


def run_polcal(
    msname="",
    field="",
    scan="",
    uvrange="",
    refant="",
    combine="",
    gaintable=[],
    gainfield=[],
    interp=[],
    n_threads=-1,
    dry_run=False,
):
    """
    Perform cross-hand phase calibration
    """
    limit_threads(n_threads=n_threads)
    from casatasks import gaincal, polcal

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    print(f"Performing relative leakage calibration on : {msname}")
    caltable_prefix = os.path.basename(msname).split(".ms")[0]
    with suppress_casa_output():
        gaincal(
            vis=msname,
            caltable=caltable_prefix + ".kcrosscal",
            field=str(field),
            scan=str(scan),
            uvrange=uvrange,
            refant=refant,
            refantmode="flex",
            solint="inf",
            combine=combine,
            gaintype="KCROSS",
            gaintable=gaintable,
            gainfield=gainfield,
            interp=interp,
        )
    if os.path.exists(caltable_prefix + ".kcrosscal"):
        gaintable.append(caltable_prefix + ".kcrosscal")
        gainfield.append(str(field))
        interp.append("")
        with suppress_casa_output():
            polcal(
                vis=msname,
                caltable=caltable_prefix + ".xfcal",
                field=str(field),
                scan=str(scan),
                uvrange=uvrange,
                refant=refant,
                solint="inf,10MHz",
                combine=combine,
                poltype="Xf",
                gaintable=gaintable,
                gainfield=gainfield,
                interp=interp,
            )
        if os.path.exists(caltable_prefix + ".xfcal"):
            gaintable.append(caltable_prefix + ".xfcal")
            gainfield.append(str(field))
            interp.append("")
            with suppress_casa_output():
                polcal(
                    vis=msname,
                    caltable=caltable_prefix + ".panglecal",
                    field=str(field),
                    scan=str(scan),
                    uvrange=uvrange,
                    refant=refant,
                    refantmode="flex",
                    solint="inf,10MHz",
                    combine="obs,scan",
                    poltype="PosAng",
                    gaintable=gaintable,
                    gainfield=gainfield,
                    interp=interp,
                )
    return (
        caltable_prefix + ".kcrosscal",
        caltable_prefix + ".xfcal",
        caltable_prefix + ".panglecal",
    )


def run_applycal(
    msname="",
    field="",
    scan="",
    applymode="",
    flagbackup=True,
    gaintable=[],
    gainfield=[],
    interp=[],
    calwt=[],
    parang=False,
    n_threads=-1,
    dry_run=False,
):
    """
    Perform apply calibration
    """
    limit_threads(n_threads=n_threads)
    from casatasks import applycal

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    print(f"Applying calibration solutions on : {msname}")
    with suppress_casa_output():
        applycal(
            vis=msname,
            field=str(field),
            scan=str(scan),
            gaintable=gaintable,
            gainfield=gainfield,
            interp=interp,
            calwt=calwt,
            applymode=applymode,
            flagbackup=flagbackup,
            parang=parang,
        )
    return


def run_postcal_flag(
    msname="",
    datacolumn="corrected",
    uvrange="",
    mode="rflag",
    n_threads=-1,
    memory_limit=-1,
    dry_run=False,
):
    """
    Perform apply calibration
    """
    limit_threads(n_threads=n_threads)
    from casatasks import flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    print(f"Post-calibration flagging on ms: {msname}")
    ncol = 3
    ####################################################
    # Check if required columns are present for residual
    ####################################################
    if datacolumn == "residual" or datacolumn == "RESIDUAL":
        modelcolumn_present = check_datacolumn_valid(msname, datacolumn="MODEL_DATA")
        corcolumn_present = check_datacolumn_valid(msname, datacolumn="CORRECTED_DATA")
        if modelcolumn_present == False or corcolumn_present == False:
            datacolumn = "corrected"
    elif datacolumn == "RESIDUAL_DATA":
        modelcolumn_present = check_datacolumn_valid(msname, datacolumn="MODEL_DATA")
        datacolumn_present = check_datacolumn_valid(msname, datacolumn="DATA")
        if modelcolumn_present == False or datacolumn_present == False:
            datacolumn = "corrected"

    #################################################
    # Whether corrected data column is present or not
    #################################################
    if datacolumn == "corrected" or datacolumn == "CORRECTED_DATA":
        corcolumn_present = check_datacolumn_valid(msname, datacolumn="CORRECTED_DATA")
        if not corcolumn_present:
            print(
                "Corrected data column is chosen for flagging, but it is not present."
            )
            return
    try:
        nchunk = get_chunk_size(msname, memory_limit=memory_limit)
        if nchunk <= 1:
            ntime = "scan"
            print("Time chunk : full scan")
        else:
            msmd = msmetadata()
            msmd.open(msname)
            scan = np.unique(msmd.scannumbers())[0]
            times = msmd.timesforspws(0)
            msmd.close()
            total_time = np.nanmax(times) - np.nanmin(times)
            timeres = np.nanmin(np.diff(times))
            ntime = float(total_time / nchunk)
            if ntime < timeres:
                ntime = timeres
            print(f"Time chunk : {ntime}s")
        with suppress_casa_output():
            flagdata(
                vis=msname,
                mode=mode,
                uvrange=uvrange,
                datacolumn=datacolumn,
                flagbackup=False,
                ntime=ntime,
            )
    except Exception as e:
        traceback.print_exc()
    return


def single_round_cal_and_flag(
    msname,
    workdir,
    cal_round,
    refant,
    uvrange,
    fluxcal_scans,
    fluxcal_fields,
    phasecal_scans,
    phasecal_fields,
    phasecal_fluxes,
    polcal_scans=[],
    polcal_fields=[],
    do_phasecal=False,
    do_leakagecal=False,
    do_polcal=False,
    do_postcal_flag=True,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Single round calibration and post-calibration flagging

    Parameters
    ----------
    msname : str
        Name of the measurement set
    workdir : str
        Work directory
    cal_round : int
        Calibration round number
    refant : str
        Reference antenna
    uvrange :str
        UV-range
    fluxcal_scans : dict
        Fluxcal scans for fluxcal fields
    fluxcal_fields : list
        Fluxcal field names
    phasecal_scans : dict
        Phasecal scans for phasecal fields
    phasecal_fields : list
        Phasecal field names
    phasecal_fluxes : dict
        Phasecal fluxes
    polcal_scans : list, optional
        Polarized calibrator scans
    polcal_fields : list, optional
        Polarized calibrator fields
    do_phasecal : bool, optional
        Perform calibration on phasecal fields
    do_leakagecal : bool, optional
        Perform leakage calibration or nor
    do_phasecal : bool, optional
        Perform calibration using polcal fields
        (Note: leakage calibration is always done using unpolarized fluxcal. This is for crossphase and polarization angle calibration)
    do_postcal_flag : bool, optional
        Peform post-calibration flagging
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    str
        Caltables
    """
    try:
        if cpu_frac > 1:
            cpu_frac = 1
        if mem_frac > 1:
            mem_frac = 1
        caltable_prefix = msname.split(".ms")[0] + "_caltable"
        msmd = msmetadata()
        msmd.open(msname)
        npol = msmd.ncorrforpol()[0]
        msmd.close()
        parang = False
        ######################################
        # Removing previous rounds caltables
        ######################################
        delay_caltable = caltable_prefix + ".kcal"
        bpass_caltable = caltable_prefix + ".bcal"
        gain_caltable = caltable_prefix + ".gcal"
        fluxscale_caltable = caltable_prefix + ".fcal"
        leakage_caltable = caltable_prefix + ".dcal"
        kcross_caltable = caltable_prefix + ".kcrosscal"
        crossphase_caltable = caltable_prefix + ".xfcal"
        pangle_caltable = caltable_prefix + ".panglecal"

        if os.path.exists(delay_caltable):
            os.system("rm -rf " + delay_caltable)
        if os.path.exists(bpass_caltable):
            os.system("rm -rf " + bpass_caltable)
        if os.path.exists(gain_caltable):
            os.system("rm -rf " + gain_caltable)
        if os.path.exists(fluxscale_caltable):
            os.system("rm -rf " + fluxscale_caltable)
        if os.path.exists(leakage_caltable):
            os.system("rm -rf " + leakage_caltable)
        if os.path.exists(kcross_caltable):
            os.system("rm -rf " + kcross_caltable)
        if os.path.exists(crossphase_caltable):
            os.system("rm -rf " + crossphase_caltable)
        if os.path.exists(pangle_caltable):
            os.system("rm -rf " + pangle_caltable)

        result = get_submsname_scans(msname)
        if result is not None:  # If multi-ms
            mslist, scans = result
        else:
            print("Please provide a multi-MS with scans partitioned.")
            return 1, []
        fluxcal_mslist = []
        phasecal_mslist = []
        polcal_mslist = []

        for fluxcal_field in fluxcal_fields:
            scan_list = fluxcal_scans[fluxcal_field]
            for scan in scan_list:
                pos = scans.index(scan)
                fluxcal_mslist.append(str(mslist[pos]))

        for phasecal_field in phasecal_fields:
            scan_list = phasecal_scans[phasecal_field]
            for scan in scan_list:
                pos = scans.index(scan)
                phasecal_mslist.append(str(mslist[pos]))

        for polcal_field in polcal_fields:
            scan_list = polcal_scans[polcal_field]
            for scan in scan_list:
                pos = scans.index(scan)
                polcal_mslist.append(str(mslist[pos]))

        #######################################
        # Calibration on fluxcal fields
        #######################################
        print("\n##############################")
        print("Calibrating fluxcal fields ....")
        print("###############################\n")
        applycal_gaintable = []
        applycal_gainfield = []
        applycal_interp = []

        if len(fluxcal_mslist) == 0 and len(phasecal_mslist) == 0:
            print("No flux calibrator or phase calibrator is present.")
            return 1, []

        ##############################
        # Delay calibration
        ##############################
        if len(fluxcal_mslist) > 0:
            delaycal_mslist = fluxcal_mslist
            #############################################
            # Memory limit
            #############################################
            ms_size_list = [get_column_size(ms) for ms in fluxcal_mslist]
            mem_limit = max(ms_size_list)
            if dask_addr is None:
                dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(delaycal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                )
            else:
                _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                    len(delaycal_mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                    only_cal=True,
                )
                os.system(f"rm -rf {dask_dir}")
                dask_client = Client(address=dask_addr)
            wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
            tasks = [
                delayed(run_delaycal)(
                    sub_msname,
                    uvrange=uvrange,
                    refant=refant,
                    solint="inf",
                    n_threads=n_threads,
                )
                for sub_msname in delaycal_mslist
            ]
            futures = dask_client.compute(tasks)
            delaycal_tables = list(dask_client.gather(futures))
            dask_client.close()
            if dask_addr is None:
                dask_cluster.close()
                os.system(f"rm -rf {dask_dir}")
            delay_caltable = merge_caltables(
                delaycal_tables, delay_caltable, keepcopy=False
            )
            if delay_caltable is not None and os.path.exists(delay_caltable):
                tb = table()
                tb.open(delay_caltable, nomodify=False)
                flag = tb.getcol("FLAG")
                flag *= False
                tb.putcol("FLAG", flag)
                tb.flush()
                tb.close()
                applycal_gaintable.append(delay_caltable)
                applycal_gainfield.append("")
                applycal_interp.append("nearest")

        ##############################
        # Bandpass calibration
        ##############################
        if len(fluxcal_mslist) > 0:
            ms_size_list = [get_column_size(ms) for ms in fluxcal_mslist]
            mem_limit = max(ms_size_list)
            if dask_addr is None:
                dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(fluxcal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                )
            else:
                _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                    len(fluxcal_mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                    only_cal=True,
                )
                os.system(f"rm -rf {dask_dir}")
                dask_client = Client(address=dask_addr)
            wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
            tasks = [
                delayed(run_bandpass)(
                    sub_msname,
                    uvrange=uvrange,
                    refant=refant,
                    solint="inf",
                    gaintable=applycal_gaintable,
                    interp=applycal_interp,
                    n_threads=n_threads,
                )
                for sub_msname in fluxcal_mslist
            ]
            futures = dask_client.compute(tasks)
            bandpass_tables = list(dask_client.gather(futures))
            dask_client.close()
            if dask_addr is None:
                dask_cluster.close()
                os.system(f"rm -rf {dask_dir}")
            bpass_caltable = merge_caltables(
                bandpass_tables, bpass_caltable, keepcopy=False
            )
            if bpass_caltable is not None and os.path.exists(bpass_caltable):
                applycal_gaintable.append(bpass_caltable)
                applycal_gainfield.append("")
                applycal_interp.append("nearestflag")
            else:
                print("Bandpass calibration is not successful.")
                return 1, []
        else:
            print("No flux calibrator is present.")
            return 1, []

        ##########################################
        # Gain calibration on fluxcal (and polcal)
        ##########################################
        gaincal_mslist = fluxcal_mslist
        if do_polcal and len(polcal_mslist) > 0 and npol == 4:
            gaincal_mslist = fluxcal_mslist + polcal_mslist
        if len(gaincal_mslist) > 0:
            ms_size_list = [get_column_size(ms) for ms in gaincal_mslist]
            mem_limit = max(ms_size_list)
            if dask_addr is None:
                dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(gaincal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                )
            else:
                _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                    len(gaincal_mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                    only_cal=True,
                )
                os.system(f"rm -rf {dask_dir}")
                dask_client = Client(address=dask_addr)
            wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
            tasks = [
                delayed(run_gaincal)(
                    sub_msname,
                    uvrange=uvrange,
                    refant=refant,
                    gaintype="T",
                    solint="1min",
                    calmode="ap",
                    gaintable=applycal_gaintable,
                    interp=applycal_interp,
                    n_threads=n_threads,
                )
                for sub_msname in gaincal_mslist
            ]
            futures = dask_client.compute(tasks)
            gain_tables = list(dask_client.gather(futures))
            dask_client.close()
            if dask_addr is None:
                dask_cluster.close()
                os.system(f"rm -rf {dask_dir}")
            gain_caltable = merge_caltables(gain_tables, gain_caltable, keepcopy=False)

        ######################################
        # Gain calibrations on phasecals
        ######################################
        if do_phasecal == False and (
            gain_caltable is not None and os.path.exists(gain_caltable)
        ):
            applycal_gaintable.append(gain_caltable)
            applycal_gainfield.append("")
            applycal_interp.append("nearest")
        else:
            print("\n##############################")
            print("Calibrating phasecal fields ....")
            print("###############################\n")
            ##############################
            # Gain calibration
            ##############################
            if len(phasecal_mslist) == 0 and (
                gain_caltable is not None and os.path.exists(gain_caltable)
            ):
                applycal_gaintable.append(gain_caltable)
                applycal_gainfield.append("")
                applycal_interp.append("nearest")
            else:
                ms_size_list = [
                    get_column_size(ms) for ms in phasecal_mslist
                ]
                mem_limit = max(ms_size_list)
                if dask_addr is None:
                    (
                        dask_client,
                        dask_cluster,
                        n_jobs,
                        n_threads,
                        mem_limit,
                        dask_dir,
                    ) = get_dask_client(
                        len(phasecal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                else:
                    _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                        len(phasecal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                        only_cal=True,
                    )
                    os.system(f"rm -rf {dask_dir}")
                    dask_client = Client(address=dask_addr)
                wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
                tasks = [
                    delayed(run_gaincal)(
                        sub_msname,
                        uvrange=uvrange,
                        refant=refant,
                        gaintype="T",
                        solint="1min",
                        calmode="ap",
                        smodel=[1, 0, 0, 0],
                        gaintable=applycal_gaintable,
                        interp=applycal_interp,
                        n_threads=n_threads,
                    )
                    for sub_msname in phasecal_mslist
                ]
                futures = dask_client.compute(tasks)
                gain_tables = list(dask_client.gather(futures))
                dask_client.close()
                if dask_addr is None:
                    dask_cluster.close()
                    os.system(f"rm -rf {dask_dir}")
                gain_caltable = merge_caltables(
                    gain_tables, gain_caltable, append=True, keepcopy=False
                )
                #################################
                # Flux scaling
                #################################
                if os.path.exists(caltable_prefix + ".listfcal"):
                    os.system("rm -rf " + caltable_prefix + ".listfcal")
                from casatasks import fluxscale

                if gain_caltable is not None and os.path.exists(gain_caltable) == False:
                    print(
                        "Gain calibration was not successful and did not produce gain caltable."
                    )
                else:
                    with suppress_casa_output():
                        fluxscale_result = fluxscale(
                            vis=msname,
                            caltable=gain_caltable,
                            fluxtable=fluxscale_caltable,
                            reference=fluxcal_fields,
                            transfer=phasecal_fields,
                        )
                    if fluxscale_caltable is not None and os.path.exists(
                        fluxscale_caltable
                    ):
                        if os.path.exists(gain_caltable):
                            os.system("rm -rf " + gain_caltable)
                        os.system(f"mv {fluxscale_caltable} {gain_caltable}")
                        applycal_gaintable.append(gain_caltable)
                        applycal_gainfield.append("")
                        applycal_interp.append("nearest")
                    msmd.open(msname)
                    phasecal_field_ids = [
                        str(msmd.fieldsforname(f)[0]) for f in phasecal_fields
                    ]
                    msmd.close()
                    for pcal in phasecal_field_ids:
                        field_name = fluxscale_result[pcal]["fieldName"]
                        catalog_flux = phasecal_fluxes[field_name]
                        result = fluxscale_result[pcal]["0"]
                        flux = result["fluxd"][0]
                        fluxerr = result["fluxdErr"][0]
                        print(f"Catalog flux: {catalog_flux} Jy")
                        print(f"Estimated flux: {flux} +/- {fluxerr} Jy")
                        print(
                            f"Difference from catalog flux: {round(abs(flux-catalog_flux),2)} Jy"
                        )
                        print("###################################\n")

        ##############################
        # Leakage calibration
        ##############################
        if do_leakagecal:
            if npol != 4:
                print(
                    "Measurement set is not full-polar. Not performing leakage calibration."
                )
            elif len(fluxcal_mslist) > 0:
                ms_size_list = [
                    get_column_size(ms) for ms in fluxcal_mslist
                ]
                mem_limit = max(ms_size_list)
                if dask_addr is None:
                    (
                        dask_client,
                        dask_cluster,
                        n_jobs,
                        n_threads,
                        mem_limit,
                        dask_dir,
                    ) = get_dask_client(
                        len(fluxcal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                else:
                    _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                        len(fluxcal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                        only_cal=True,
                    )
                    os.system(f"rm -rf {dask_dir}")
                    dask_client = Client(address=dask_addr)
                wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
                tasks = [
                    delayed(run_leakagecal)(
                        sub_msname,
                        uvrange=uvrange,
                        refant=refant,
                        gaintable=applycal_gaintable,
                        gainfield=["", "", ",".join(fluxcal_fields)],
                        interp=applycal_interp,
                        n_threads=n_threads,
                    )
                    for sub_msname in fluxcal_mslist
                ]
                futures = dask_client.compute(tasks)
                leakage_tables = list(dask_client.gather(futures))
                dask_client.close()
                if dask_addr is None:
                    dask_cluster.close()
                    os.system(f"rm -rf {dask_dir}")
                leakage_caltable = merge_caltables(
                    leakage_tables, leakage_caltable, keepcopy=False
                )
                if leakage_caltable is not None and os.path.exists(leakage_caltable):
                    applycal_gaintable.append(leakage_caltable)
                    applycal_gainfield.append("")
                    applycal_interp.append("nearest,nearestflag")
                    if not parang:
                        parang = True

        ########################################
        # Calibration using polarized calibrator
        ########################################
        if do_polcal and do_leakagecal:
            if len(polcal_mslist) == 0:
                print("No polarized calibrator fields are present.")
            elif npol != 4:
                print("Measurement set is not full-polar.")
            elif os.path.exists(leakage_caltable) == False:
                print("Leakage solutions are not present.")
            else:
                ms_size_list = [get_column_size(ms) for ms in polcal_mslist]
                mem_limit = max(ms_size_list)
                if dask_addr is None:
                    (
                        dask_client,
                        dask_cluster,
                        n_jobs,
                        n_threads,
                        mem_limit,
                        dask_dir,
                    ) = get_dask_client(
                        len(polcal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                else:
                    _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                        len(polcal_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                        only_cal=True,
                    )
                    os.system(f"rm -rf {dask_dir}")
                    dask_client = Client(address=dask_addr)
                wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
                tasks = [
                    delayed(run_polcal)(
                        sub_msname,
                        uvrange=uvrange,
                        refant=refant,
                        solint="inf",
                        gaintable=applycal_gaintable,
                        gainfield=[
                            "",
                            "",
                            ",".join(polcal_fields),
                            ",".join(fluxcal_fields),
                        ],
                        interp=applycal_interp,
                        n_threads=n_threads,
                    )
                    for sub_msname in polcal_mslist
                ]
                futures = dask_client.compute(tasks)
                results = list(dask_client.gather(futures))
                dask_client.close()
                if dask_addr is None:
                    dask_cluster.close()
                    os.system(f"rm -rf {dask_dir}")
                kcross_tables = []
                crossphase_tables = []
                pangle_tables = []
                for r in results:
                    kcross_tables.append(r[0])
                    if len(r) > 1:
                        crossphase_tables.append(r[1])
                    if len(r) > 2:
                        pangle_tables.append(r[2])
                kcross_caltable = merge_caltables(
                    kcross_tables, kcross_caltable, keepcopy=False
                )
                crossphase_caltable = merge_caltables(
                    crossphase_tables, crossphase_caltable, keepcopy=False
                )
                pangle_caltable = merge_caltables(
                    pangle_tables, pangle_caltable, keepcopy=False
                )

                if kcross_caltable is not None and os.path.exists(kcross_caltable):
                    applycal_gaintable.append(kcross_caltable)
                    applycal_gainfield.append("")
                    applycal_interp.append("nearest")
                    if not parang:
                        parang = True
                    if crossphase_caltable is not None and os.path.exists(
                        crossphase_caltable
                    ):
                        applycal_gaintable.append(crossphase_caltable)
                        applycal_gainfield.append("")
                        applycal_interp.append("nearest,nearestflag")
                        if pangle_caltable is not None and os.path.exists(
                            pangle_caltable
                        ):
                            applycal_gaintable.append(pangle_caltable)
                            applycal_gainfield.append("")
                            applycal_interp.append("nearest,nearestflag")
                        else:
                            print(
                                "Absolute polarization angle calibration could not be done."
                            )
                    else:
                        print(
                            "Crosshand phase and absolute polarization angle calibration could not be done."
                        )
                else:
                    print(
                        "Crosshand delay, crosshand phase and absolute polarization angle calibration could not be done."
                    )

        ##############################
        # Apply calibration
        ##############################
        all_mslist = copy.deepcopy(fluxcal_mslist)
        if len(phasecal_mslist) > 0:
            all_mslist += phasecal_mslist
        if len(polcal_mslist) > 0:
            all_mslist += polcal_mslist
        if len(all_mslist) > 0:
            do_flag_backup(msname, flagtype="applycal")
            ms_size_list = [get_column_size(ms) for ms in all_mslist]
            mem_limit = max(ms_size_list)
            if dask_addr is None:
                dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(all_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                )
            else:
                _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                    len(all_mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                    only_cal=True,
                )
                os.system(f"rm -rf {dask_dir}")
                dask_client = Client(address=dask_addr)
            wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
            tasks = [
                delayed(run_applycal)(
                    sub_msname,
                    flagbackup=False,
                    gaintable=applycal_gaintable,
                    gainfield=applycal_gainfield,
                    interp=applycal_interp,
                    calwt=[False] * len(applycal_gainfield),
                    parang=parang,
                    n_threads=n_threads,
                )
                for sub_msname in all_mslist
            ]
            futures = dask_client.compute(tasks)
            results = list(dask_client.gather(futures))
            dask_client.close()
            if dask_addr is None:
                dask_cluster.close()
                os.system(f"rm -rf {dask_dir}")

        ##############################
        # Post calibration flagging
        ##############################
        if do_postcal_flag and len(all_mslist) > 0:
            do_flag_backup(msname, flagtype="flagdata")
            ms_size_list = [get_column_size(ms) for ms in all_mslist]
            mem_limit = max(ms_size_list)
            if dask_addr is None:
                dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                    get_dask_client(
                        len(all_mslist),
                        dask_dir=workdir,
                        cpu_frac=cpu_frac,
                        mem_frac=mem_frac,
                        min_mem_per_job=mem_limit,
                    )
                )
            else:
                _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                    len(all_mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                    only_cal=True,
                )
                os.system(f"rm -rf {dask_dir}")
                dask_client = Client(address=dask_addr)
            wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
            tasks = []
            if len(all_mslist) > 0:
                tasks = []
                for sub_msname in all_mslist:
                    if sub_msname in fluxcal_mslist:
                        datacolumn = "residual"
                    else:
                        datacolumn = "corrected"
                    tasks.append(
                        delayed(run_postcal_flag)(
                            sub_msname,
                            datacolumn=datacolumn,
                            uvrange=uvrange,
                            mode="rflag",
                            n_threads=n_threads,
                            memory_limit=mem_limit,
                        )
                    )
            futures = dask_client.compute(tasks)
            results = list(dask_client.gather(futures))
            dask_client.close()
            if dask_addr is None:
                dask_cluster.close()
                os.system(f"rm -rf {dask_dir}")

        ###############################
        # Finished calibration round
        ###############################
        delay_caltable = (
            delay_caltable
            if (delay_caltable is not None and os.path.exists(delay_caltable))
            else None
        )
        bpass_caltable = (
            bpass_caltable
            if (bpass_caltable is not None and os.path.exists(bpass_caltable))
            else None
        )
        gain_caltable = (
            gain_caltable
            if (gain_caltable is not None and os.path.exists(gain_caltable))
            else None
        )
        leakage_caltable = (
            leakage_caltable
            if (leakage_caltable is not None and os.path.exists(leakage_caltable))
            else None
        )
        kcross_caltable = (
            kcross_caltable
            if (kcross_caltable is not None and os.path.exists(kcross_caltable))
            else None
        )
        crossphase_caltable = (
            crossphase_caltable
            if (crossphase_caltable is not None and os.path.exists(crossphase_caltable))
            else None
        )
        pangle_caltable = (
            pangle_caltable
            if (pangle_caltable is not None and os.path.exists(pangle_caltable))
            else None
        )
        return 0, [
            delay_caltable,
            bpass_caltable,
            gain_caltable,
            leakage_caltable,
            kcross_caltable,
            crossphase_caltable,
            pangle_caltable,
        ]
    except Exception as e:
        traceback.print_exc()
        return 1, []


def run_basic_cal_rounds(
    msname,
    workdir,
    refant="",
    uvrange="",
    keep_backup=False,
    perform_polcal=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Perform basic calibration rounds

    Parameters
    ----------
    msname : str
        Name of the measurement set
    workdir : str
        Warking directory
    refant : str, optional
        Reference antenna
    uvrange : str, optional
        UV-range
    perform_polcal : bool, optional
        Perform polarization calibration for fullpolar data
    keep_backup : bool, optional
        Keep backup of ms after each calibration rounds
    cpu_frac : float, options
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    list
        Caltables
    """
    start_time = time.time()
    try:
        os.chdir(workdir)
        print(f"Measurement set : {msname}")
        print("Extracting metadata from measurement set ....")
        correct_missing_col_subms(msname)
        fluxcal_fields, fluxcal_scans = get_fluxcals(msname)
        polcal_fields, polcal_scans = get_polcals(msname)
        phasecal_fields, phasecal_scans, phasecal_fluxes = get_phasecals(msname)
        msmd = msmetadata()
        msmd.open(msname)
        npol = msmd.ncorrforpol()[0]
        msmd.close()
        if npol == 4 or len(polcal_fields) > 0 or len(phasecal_fields) > 0:
            n_rounds = 3
        else:
            n_rounds = 2
        print(f"Total calibration rounds: {n_rounds}")
        do_phasecal = False
        do_polcal = False
        do_leakagecal = False
        do_postcal_flag = True
        ###################################################
        # Determining what calibrations will be done or not
        ###################################################
        if len(phasecal_fields) > 0:
            perform_phasecal = True
        else:
            perform_phasecal = False
        if perform_polcal and npol == 4:
            perform_leakagecal = True
            if len(polcal_fields) > 0:
                perform_polcal = (
                    True  # Leakage calibration is done using unpolarized fluxcal
                )
            else:
                perform_polcal = False
        else:  # If not a full polar ms
            print(
                "Measurement set is not full-polar. Do not performing any polarization calibration."
            )
            perform_leakagecal = False
            perform_polcal = False
        #####################################################
        if refant == "":
            refant = get_refant(msname)
        if uvrange == "":
            uvrange = ">200lambda"
        print("#########################################")
        print(f"Using UV-range for calibration: {uvrange}")
        print("#########################################")
        for cal_round in range(1, n_rounds + 1):
            print("\n#################################")
            print(f"Calibration round: {cal_round}")
            print("#################################\n")
            if cal_round == n_rounds:
                do_postcal_flag = False
            if cal_round > 1:
                if perform_phasecal:
                    do_phasecal = True
                if perform_polcal:
                    do_polcal = True
                if perform_leakagecal:
                    do_leakagecal = True
            msg, caltables = single_round_cal_and_flag(
                msname,
                workdir,
                cal_round,
                refant,
                uvrange,
                fluxcal_scans,
                fluxcal_fields,
                phasecal_scans,
                phasecal_fields,
                phasecal_fluxes,
                do_phasecal=do_phasecal,
                do_leakagecal=do_leakagecal,
                do_polcal=do_polcal,
                do_postcal_flag=do_postcal_flag,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
            if keep_backup:
                print(f"Backup directory: {workdir}/backup")
                os.makedirs(workdir + "/backup", exist_ok=True)
                for caltable in caltables:
                    if caltable is not None and os.path.exists(caltable):
                        cal_ext = os.path.basename(caltable).split(".")[-1]
                        outputname = (
                            workdir
                            + "/backup/"
                            + os.path.basename(caltable).split(f".{cal_ext}")[0]
                            + "_round_"
                            + str(cal_round)
                            + f".{cal_ext}"
                        )
                        os.system("cp -r " + caltable + " " + outputname)
            if msg == 1:
                print("##################")
                print("Basic calibration is not successful.")
                print(f"Total time taken : {time.time() - start_time}")
                print("##################\n")
                return 1, []
        print("##################")
        print("Basic calibration is done successfully.")
        print(f"Total time taken : {time.time() - start_time}")
        print("##################\n")
        return 0, caltables
    except Exception as e:
        traceback.print_exc()
        print("##################")
        print(f"Total time taken : {time.time() - start_time}")
        print("##################\n")
        return 1, []


def main(
    msname,
    workdir,
    caldir,
    refant="1",
    uvrange="",
    perform_polcal=False,
    keep_backup=False,
    start_remote_log=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid=0,
    dask_addr=None,
):
    """
    Main function to perform basic calibration

    Parameters
    ----------
    msname : str
        Measurement set
    workdir : str
        Work directory
    caldir : str
        Caltables directory
    refant : str, optional
        Reference antenna
    uvrange : str, optional
        UV-range
    perform_polcal : bool, optional
        Perform polarization calibration
    start_remote_log : bool, optional
        Start logging to remote logger or not
    keep_backup : bool, optional
        Keep backup
    cpu_frac : float, optional
        CPU fraction
    mem_frac : float, optional
        Memory fraction
    logfile : str, optional
        Log file name
    jobid : str, optional
        Pipeline Job ID
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(msname)) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    if caldir == "" or not os.path.exists(caldir):
        caldir = f"{workdir}/caltables"
    os.makedirs(caldir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "basic_cal", logfile, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")

    try:
        if msname != "" and os.path.exists(msname):
            print("\n###################################")
            print("Starting initial calibration.")
            print("###################################\n")
            msg, caltables = run_basic_cal_rounds(
                msname,
                workdir,
                refant=refant,
                uvrange=uvrange,
                perform_polcal=perform_polcal,
                keep_backup=keep_backup,
                cpu_frac=float(cpu_frac),
                mem_frac=float(mem_frac),
                dask_addr=dask_addr,
            )

            for caltable in caltables:
                if caltable is not None and os.path.exists(caltable):
                    dest = caldir + "/" + os.path.basename(caltable)
                    if os.path.exists(dest):
                        os.system("rm -rf " + dest)
                    os.system("mv " + caltable + " " + caldir)
        else:
            print("Please provide a valid measurement set.\n")
            msg = 1
    except Exception as e:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        drop_cache(msname)
        drop_cache(workdir)
        drop_cache(caldir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Basic calibration using calibrators",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "msname",
        type=str,
        help="Name of measurement set (required positional argument)",
    )
    basic_args.add_argument(
        "--workdir",
        type=str,
        default="",
        required=True,
        help="Working directory for calibration outputs (default: auto-created next to MS)",
    )
    basic_args.add_argument(
        "--caldir",
        type=str,
        default="",
        required=True,
        help="Caltables directory (default: auto-created in the workdir MS)",
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced calibration parameters\n###################"
    )
    adv_args.add_argument("--refant", type=str, default="1", help="Reference antenna")
    adv_args.add_argument(
        "--uvrange",
        type=str,
        default="",
        help="UV range for calibration (e.g. '>100lambda')",
    )
    adv_args.add_argument(
        "--perform_polcal", action="store_true", help="Enable polarization calibration"
    )
    adv_args.add_argument(
        "--keep_backup",
        action="store_true",
        help="Keep backup of measurement set after each calibration round",
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac", type=float, default=0.8, help="CPU fraction to use"
    )
    hard_args.add_argument(
        "--mem_frac", type=float, default=0.8, help="Memory fraction to use"
    )
    hard_args.add_argument(
        "--logfile", type=str, default=None, help="Optional path to log file"
    )
    hard_args.add_argument(
        "--jobid", type=str, default="0", help="Job ID for logging and PID tracking"
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        return 0

    args = parser.parse_args()

    msg = main(
        args.msname,
        args.workdir,
        args.caldir,
        refant=args.refant,
        uvrange=args.uvrange,
        perform_polcal=args.perform_polcal,
        keep_backup=args.keep_backup,
        start_remote_log=args.start_remote_log,
        cpu_frac=float(args.cpu_frac),
        mem_frac=float(args.mem_frac),
        logfile=args.logfile,
        jobid=args.jobid,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nBasic calibration is finished.\n###################\n"
    )
    os._exit(result)



import types
import numpy as np
import glob
import os
from casatools import msmetadata, ms as casamstool, table
from .basic_utils import *

##################################
# Imaging related
##################################
def calc_sun_dia(freqMHz):
    """
    Function to calculate the diameter of the Sun at a given frequency (White 2016)

    Parameters
    ----------
    freq : float
        Frequency in MHz

    Returns
    -------
    float
        Diameter of the Sun in arcmin
    """
    freqGHz = freqMHz / 10**3  # Convert in GHz
    dia = 32 + (2.2 * (freqGHz) ** (-0.6))
    return round(dia, 2)


def calc_maxuv(msname, chan_number=-1):
    """
    Calculate maximum UV

    Parameters
    ----------
    msname : str
        Name of the measurement set
    chan_number : int, optional
        Channel number

    Returns
    -------
    float
        Maximum UV in meter
    float
        Maximum UV in wavelength
    """
    msmd = msmetadata()
    msmd.open(msname)
    freq = msmd.chanfreqs(0)[chan_number]
    wavelength = 299792458.0 / (freq)
    msmd.close()
    msmd.done()
    tb = table()
    tb.open(msname)
    uvw = tb.getcol("UVW")
    tb.close()
    u, v, w = [uvw[i, :] for i in range(3)]
    uv = np.sqrt(u**2 + v**2)
    uv[uv == 0] = np.nan
    maxuv = np.nanmax(uv)
    return round(float(maxuv), 2), round(float(maxuv / wavelength), 2)


def calc_minuv(msname, chan_number=-1):
    """
    Calculate minimum UV

    Parameters
    ----------
    msname : str
        Name of the measurement set
    chan_number : int, optional
        Channel number

    Returns
    -------
    float
        Minimum UV in meter
    float
        Minimum UV in wavelength
    """
    msmd = msmetadata()
    msmd.open(msname)
    freq = msmd.chanfreqs(0)[chan_number]
    wavelength = 299792458.0 / (freq)
    msmd.close()
    msmd.done()
    tb = table()
    tb.open(msname)
    uvw = tb.getcol("UVW")
    tb.close()
    u, v, w = [uvw[i, :] for i in range(3)]
    uv = np.sqrt(u**2 + v**2)
    uv[uv == 0] = np.nan
    minuv = np.nanmin(uv)
    return round(float(minuv), 2), round(float(minuv / wavelength), 2)


def calc_field_of_view(msname, FWHM=True):
    """
    Calculate optimum field of view in arcsec.

    Parameters
    ----------
    msname : str
        Measurement set name
    FWHM : bool, optional
        Upto FWHM, otherwise upto first null

    Returns
    -------
    float
        Field of view in arcsec
    """
    msmd = msmetadata()
    msmd.open(msname)
    freq = msmd.chanfreqs(0)[0]
    msmd.close()
    tb = table()
    tb.open(msname + "/ANTENNA")
    dish_dia = np.nanmin(tb.getcol("DISH_DIAMETER"))
    tb.close()
    wavelength = 299792458.0 / freq
    if FWHM:
        FOV = 1.22 * wavelength / dish_dia
    else:
        FOV = 2.04 * wavelength / dish_dia
    fov_arcsec = np.rad2deg(FOV) * 3600  # In arcsecs
    return round(float(fov_arcsec), 2)


def get_optimal_image_interval(
    msname,
    temporal_tol_factor=0.1,
    spectral_tol_factor=0.1,
    chan_range="",
    timestamp_range="",
    max_nchan=-1,
    max_ntime=-1,
):
    """
    Get optimal image spectral temporal interval such that total flux max-median in each chunk is within tolerance limit

    Parameters
    ----------
    msname : str
        Name of the measurement set
    temporal_tol_factor : float, optional
        Tolerance factor for temporal variation (default : 0.1, 10%)
    spectral_tol_factor : float, optional
        Tolerance factor for spectral variation (default : 0.1, 10%)
    chan_range : str, optional
        Channel range
    timestamp_range : str, optional
        Timestamp range
    max_nchan : int, optional
        Maxmium number of spectral chunk
    max_ntime : int, optional
        Maximum number of temporal chunk

    Returns
    -------
    int
        Number of time intervals to average
    int
        Number of channels to averages
    """

    def is_valid_chunk(chunk, tolerance):
        mean_flux = np.nanmedian(chunk)
        if mean_flux == 0:
            return False
        return (np.nanmax(chunk) - np.nanmin(chunk)) / mean_flux <= tolerance

    def find_max_valid_chunk_length(fluxes, tolerance):
        n = len(fluxes)
        for window in range(n, 1, -1):  # Try from largest to smallest
            valid = True
            for start in range(0, n, window):
                end = min(start + window, n)
                chunk = fluxes[start:end]
                if len(chunk) < window:  # Optionally require full window
                    valid = False
                    break
                if not is_valid_chunk(chunk, tolerance):
                    valid = False
                    break
            if valid:
                return window  # Return the largest valid window
        return 1  # Minimum chunk size is 1 if nothing else is valid

    tb = table()
    mstool = casamstool()
    msmd = msmetadata()
    msmd.open(msname)
    nchan = msmd.nchan(0)
    times = msmd.timesforspws(0)
    ntime = len(times)
    del times
    msmd.close()
    tb.open(msname)
    u, v, w = tb.getcol("UVW")
    tb.close()
    uvdist = np.sort(np.unique(np.sqrt(u**2 + v**2)))
    mstool.open(msname)
    if uvdist[0] == 0.0:
        mstool.select({"uvdist": [0.0, 0.0]})
    else:
        mstool.select({"antenna1": 0, "antenna2": 1})
    data_and_flag = mstool.getdata(["DATA", "FLAG"], ifraxis=True)
    data = data_and_flag["data"]
    flag = data_and_flag["flag"]
    data[flag] = np.nan
    mstool.close()
    if chan_range != "":
        start_chan = int(chan_range.split(",")[0])
        end_chan = int(chan_range.split(",")[-1])
        spectra = np.nanmedian(data[:, start_chan:end_chan, ...], axis=(0, 2, 3))
    else:
        spectra = np.nanmedian(data, axis=(0, 2, 3))
    if timestamp_range != "":
        t_start = int(timestamp_range.split(",")[0])
        t_end = int(timestamp_range.split(",")[-1])
        t_series = np.nanmedian(data[..., t_start:t_end], axis=(0, 1, 2))
    else:
        t_series = np.nanmedian(data, axis=(0, 1, 2))
    t_series = t_series[t_series != 0]
    spectra = spectra[spectra != 0]
    t_chunksize = find_max_valid_chunk_length(t_series, temporal_tol_factor)
    f_chunksize = find_max_valid_chunk_length(spectra, spectral_tol_factor)
    n_time_interval = int(len(t_series) / t_chunksize)
    n_spectral_interval = int(len(spectra) / f_chunksize)
    if max_nchan > 0 and n_spectral_interval > max_nchan:
        n_spectral_interval = max_nchan
    if max_ntime > 0 and n_time_interval > max_ntime:
        n_time_interval = max_ntime
    return n_time_interval, n_spectral_interval


def calc_psf(msname, chan_number=-1):
    """
    Function to calculate PSF size in arcsec

    Parameters
    ----------
    msname : str
        Name of the measurement set
    chan_number : int, optional
        Channel number

    Returns
    -------
    float
            PSF size in arcsec
    """
    maxuv_m, maxuv_l = calc_maxuv(msname, chan_number=chan_number)
    psf = np.rad2deg(1.2 / maxuv_l) * 3600.0  # In arcsec
    return round(float(psf), 2)


def calc_npix_in_psf(weight, robust=0.0):
    """
    Calculate number of pixels in a PSF (could be in fraction)

    Parameters
    ----------
    weight : str
        Image weighting scheme
    robust : float, optional
        Briggs weighting robust parameter (-1,1)

    Returns
    -------
    float
        Number of pixels in a PSF
    """
    if weight.upper() == "NATURAL":
        npix = 3.0
    elif weight.upper() == "UNIFORM":
        npix = 5.0
    else:  # -1 to +1, uniform to natural
        robust = np.clip(robust, -1.0, 1.0)
        npix = 5.0 - ((robust + 1.0) / 2.0) * (5.0 - 3.0)
    return round(npix, 1)


def calc_cellsize(msname, num_pixel_in_psf):
    """
    Calculate pixel size in arcsec

    Parameters
    ----------
    msname : str
        Name of the measurement set
    num_pixel_in_psf : float
        Number of pixels in one PSF

    Returns
    -------
    int
        Pixel size in arcsec
    """
    psf = calc_psf(msname)
    pixel = round(psf / num_pixel_in_psf, 1)
    return pixel


def calc_multiscale_scales(msname, num_pixel_in_psf, chan_number=-1, max_scale=16):
    """
    Calculate multiscale scales

    Parameters
    ----------
    msname : str
        Name of the measurement set
    num_pixel_in_psf : float
            Number of pixels in one PSF
    max_scale : float, optional
        Maximum scale in arcmin

    Returns
    -------
    list
            Multiscale scales in pixel units
    """
    psf = calc_psf(msname, chan_number=chan_number)
    minuv, minuv_l = calc_minuv(msname, chan_number=chan_number)
    max_interferometric_scale = (
        0.5 * np.rad2deg(1.0 / minuv_l) * 60.0
    )  # In arcmin, half of maximum scale
    max_interferometric_scale = min(max_scale, max_interferometric_scale)
    max_scale_pixel = int((max_interferometric_scale * 60.0) / (psf / num_pixel_in_psf))
    multiscale_scales = [0]
    current_scale = num_pixel_in_psf
    while True:
        current_scale = current_scale * 2
        if current_scale >= max_scale_pixel:
            break
        multiscale_scales.append(current_scale)
    return multiscale_scales


def get_multiscale_bias(freq, bias_min=0.6, bias_max=0.9, minfreq=1015, maxfreq=1670):
    """
    Get frequency dependent multiscale bias

    Parameters
    ----------
    freq : float
        Frequency in MHz
    bias_min : float, optional
        Minimum bias at minimum L-band frequency
    bias_max : float, optional
        Maximum bias at maximum L-band frequency
    minfreq : float, optional
        Minimum frequency range in MHz
    maxfreq : float, optional
        Maximum frequency range in MHz

    Returns
    -------
    float
        Multiscale bias patrameter
    """
    if freq <= minfreq:
        return bias_min
    elif freq >= maxfreq:
        return bias_max
    else:
        freq_min = 1015
        freq_max = 1670
        logf = np.log10(freq)
        logf_min = np.log10(freq_min)
        logf_max = np.log10(freq_max)
        frac = (logf - logf_min) / (logf_max - logf_min)
        return round(
            np.clip(bias_min + frac * (bias_max - bias_min), bias_min, bias_max), 3
        )


# Expose functions and classes
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]



import logging
import psutil
import dask
import numpy as np
import argparse
import traceback
import warnings
import copy
import time
import glob
import sys
import os
from casatasks import casalog
from casatools import table
from dask import delayed
from dask.distributed import Client
from scipy.interpolate import CubicSpline
from scipy.ndimage import gaussian_filter1d
from scipy.interpolate import interp1d
from meersolar.utils import *
from meersolar.meerpipeline.flagging import single_ms_flag

logging.getLogger("distributed").setLevel(logging.WARNING)


try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def interpolate_nans(data):
    """Linearly interpolate NaNs in 1D array."""
    nans = np.isnan(data)
    if np.all(nans):
        raise ValueError("All values are NaN.")
    x = np.arange(len(data))
    interp_func = interp1d(
        x[~nans],
        data[~nans],
        kind="linear",
        bounds_error=False,
        fill_value="extrapolate",
    )
    return interp_func(x)


def filter_outliers(data, threshold=5, max_iter=3):
    """
    Filter outliers and perform cubic spline fitting

    Parameters
    ----------
    y : numpy.array
        Y values
    threshold : float
        Threshold of filtering
    max_iter : int
        Maximum number of iterations

    Returns
    -------
    numpy.array
        Clean Y-values
    """
    for c in range(max_iter):
        data = np.asarray(data, dtype=np.float64)
        original_nan_mask = np.isnan(data)
        # Interpolate NaNs for smoothing
        interpolated_data = interpolate_nans(data)
        # Apply Gaussian smoothing
        smoothed = gaussian_filter1d(
            interpolated_data, sigma=threshold, truncate=3 * threshold
        )
        # Compute residuals and std only on valid original data
        residuals = data - smoothed
        valid_mask = ~original_nan_mask
        std_dev = np.std(residuals[valid_mask])
        # Detect outliers
        outlier_mask = np.abs(residuals) > threshold * std_dev
        combined_mask = valid_mask & ~outlier_mask
        # Replace outliers with NaN
        filtered_data = np.where(combined_mask, data, np.nan)
        data = copy.deepcopy(filtered_data)
    return filtered_data


def scale_bandpass(bandpass_table, att_tables):
    """
    Scale a bandpass calibration table using attenuation data.

    Parameters
    ----------
    bandpass_table : str
        Input bandpass calibration table.
    att_tables : list
        Attenuation NumPy .npy files containing attenuation frequency and values for different scans.

    Returns
    -------
    str
        Name of the output table.
    """
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    if isinstance(att_tables, str):
        att_tables = [att_tables]
    if len(att_tables) == 0:
        print("No attenuation table is provided.")
        return
    scan_att_array = []
    for att_table in att_tables:
        results = np.load(att_table, allow_pickle=True)
        scan, freqs, att_values, flag_ants, att_array = results
        scan_att_array.append(att_values)
    scan_att_array = np.array(scan_att_array)
    att_values = np.nanmedian(scan_att_array, axis=0)  # Median across scans
    output_table = bandpass_table.split(".bcal")[0] + "_att.bcal"
    tb = table()
    tb.open(f"{bandpass_table}/SPECTRAL_WINDOW")
    caltable_freqs = tb.getcol("CHAN_FREQ").flatten()
    tb.close()
    # Prepare output table
    if os.path.exists(output_table):
        os.system(f"rm -rf {output_table}")
    os.system(f"cp -r {bandpass_table} {output_table}")
    tb.open(output_table, nomodify=False)
    gain = tb.getcol("CPARAM")
    flag = tb.getcol("FLAG")
    for i in range(att_values.shape[0]):
        att = filter_outliers(att_values[i])
        valid = ~np.isnan(att)
        best_fit, best_std = None, np.inf
        for deg in range(3, 9):
            coeffs = np.polyfit(freqs[valid], att[valid], deg)
            interp_func = np.poly1d(coeffs)
            interp_att = interp_func(caltable_freqs)
            residuals = att_values[i] - interp_att
            new_std = np.nanstd(residuals[~np.isnan(att_values[i])])
            if new_std >= best_std:
                break
            best_std = new_std
            best_fit = interp_att
        # Limit interpolation to valid frequency range
        nanpos = np.where(~np.isnan(att_values[i]))[0]
        minpos, maxpos = np.nanmin(nanpos), np.nanmax(nanpos)
        best_fit[:minpos] = best_fit[maxpos:] = np.nan
        # Broadcast to CPARAM shape
        interp_scaled = np.sqrt(best_fit)
        # Apply scaling
        gain[i, ...] *= interp_scaled[..., None]
    gain[flag] = 1.0
    tb.putcol("CPARAM", gain)
    tb.flush()
    tb.close()
    return output_table


def applysol(
    msname="",
    gaintable=[],
    gainfield=[],
    interp=[],
    parang=False,
    applymode="calflag",
    overwrite_datacolumn=False,
    n_threads=-1,
    memory_limit=-1,
    force_apply=False,
    soltype="basic",
    do_post_flag=False,
    dry_run=False,
):
    """
    Apply flux calibrated and attenuation calibrated solutions

    Parameters
    ----------
    msname : str
        Measurement set
    gaintable : list, optional
        Caltable list
    gainfield : list, optional
        Gain field list
    interp : list, optional
        Gain interpolation
    parang : bool, optional
        Parallactic angle apply or not
    applymode : str, optional
        Apply mode
    overwrite_datacolumn : bool, optional
        Overwrite data column with corrected solutions
    n_threads : int, optional
        Number of OpenMP threads
    memory_limit : float, optional
        Memory limit in GB
    force_apply : bool, optional
        Force to apply solutions if it is already applied
    soltype : str, optional
        Solution type
    do_post_flag : bool, optional
        Do post calibration flagging

    Returns
    -------
    int
        Success message
    """
    limit_threads(n_threads=n_threads)
    from casatasks import applycal, flagdata, split, clearcal

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    print(
        f"Applying solutions on ms: {os.path.basename(msname)} from caltables: {','.join([os.path.basename(i) for i in gaintable])}\n"
    )
    if soltype == "basic":
        check_file = "/.applied_sol"
    else:
        check_file = "/.applied_selfcalsol"
    try:
        if os.path.exists(msname + check_file) and force_apply == False:
            print("Solutions are already applied.")
            return 0
        else:
            if os.path.exists(msname + check_file) and force_apply:
                with suppress_casa_output():
                    clearcal(vis=msname)
                    flagdata(vis=msname, mode="unflag", spw="0", flagbackup=False)
                if os.path.exists(msname + ".flagversions"):
                    os.system("rm -rf " + msname + ".flagversions")
            with suppress_casa_output():
                applycal(
                    vis=msname,
                    gaintable=gaintable,
                    gainfield=gainfield,
                    applymode=applymode,
                    interp=interp,
                    calwt=[False] * len(gaintable),
                    parang=parang,
                    flagbackup=False,
                )
        if overwrite_datacolumn:
            print(f"Spliting corrected data for ms: {msname}.")
            outputvis = msname.split(".ms")[0] + "_cor.ms"
            if os.path.exists(outputvis):
                os.system(f"rm -rf {outputvis}")
            touch_file_names = glob.glob(f"{msname}/.*")
            if len(touch_file_names) > 0:
                touch_file_names = [os.path.basename(f) for f in touch_file_names]
            with suppress_casa_output():
                split(vis=msname, outputvis=outputvis, datacolumn="corrected")
            if os.path.exists(outputvis):
                os.system(f"rm -rf {msname} {msname}.flagversions")
                os.system(f"mv {outputvis} {msname}")
            for t in touch_file_names:
                os.system(f"touch {msname}/{t}")
        if do_post_flag:
            print(f"Post calibration flagging on: {msname}")
            if overwrite_datacolumn:
                datacolumn = "data"
            else:
                datacolumn = "corrected"
            single_ms_flag(
                msname=msname,
                datacolumn=datacolumn,
                use_tfcrop=True,
                use_rflag=False,
                flagdimension="freq",
                flag_autocorr=False,
                n_threads=n_threads,
                memory_limit=memory_limit,
            )
        os.system("touch " + msname + check_file)
        return 0
    except Exception as e:
        traceback.print_exc()
        return 1


def run_all_applysol(
    mslist,
    workdir,
    caldir,
    use_only_bandpass=False,
    overwrite_datacolumn=False,
    applymode="calflag",
    force_apply=False,
    do_post_flag=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    dask_addr=None,
):
    """
    Apply basic-calibration solutions on all target scans

    Parameters
    ----------
    mslist : str
        Measurement set list
    workdir : str
        Working directory
    caldir : str
        Calibration directory
    use_only_bandpass : bool, optional
        Use only bandpass solutions
    overwrite_datacolumn : bool, optional
        Overwrite data column or not
    applymode : str, optional
        Apply mode
    force_apply : bool, optional
        Force to apply solutions even already applied
    do_post_flag : bool, optional
        Do post calibration flagging
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    --------
    list
        Calibrated target scans
    """
    start_time = time.time()
    try:
        os.chdir(workdir)
        mslist = np.unique(mslist).tolist()
        parang = False
        os.system("rm -rf " + caldir + "/*scan*.bcal")
        att_caltables = glob.glob(caldir + "/*_attval_scan_*.npy")
        bandpass_table = glob.glob(caldir + "/calibrator_caltable.bcal")
        delay_table = glob.glob(caldir + "/calibrator_caltable.kcal")
        gain_table = glob.glob(caldir + "/calibrator_caltable.gcal")
        leakage_table = glob.glob(caldir + "/calibrator_caltable.dcal")
        if len(leakage_table) > 0:
            parang = True
            kcross_table = glob.glob(caldir + "/calibrator_caltable.kcrosscal")
            crossphase_table = glob.glob(caldir + "/calibrator_caltable.xfcal")
            pangle_table = glob.glob(caldir + "/calibrator_caltable.panglecal")
        else:
            print(f"No polarization leakage calibration table is present in : {caldir}")
            kcross_table = []
            crossphase_table = []
            pangle_table = []

        gaintable = []
        if len(bandpass_table) == 0:
            print(f"No bandpass table is present in calibration directory : {caldir}.")
            return []
        if len(gain_table) == 0:
            print(
                f"No time-dependent gaintable is present in calibration directory : {caldir}. Applying only bandpass solutions."
            )
            use_only_bandpass = True

        ################################
        # Scale bandpass for attenuators
        ################################
        if len(att_caltables) == 0:
            print(
                "No attenuation table is present. Bandpass is not scaled for attenuation."
            )
            scaled_bandpass_table = ""
        else:
            scaled_bandpass_table = scale_bandpass(bandpass_table[0], att_caltables)

        ###############################
        # Arranging applycal
        ###############################
        if len(delay_table) > 0:
            gaintable += delay_table
        if len(gain_table) > 0 and use_only_bandpass == False:
            gaintable += gain_table
        if len(leakage_table) > 0:
            gaintable += leakage_table
            if len(kcross_table) > 0:
                gaintable += kcross_table
            if len(crossphase_table) > 0:
                gaintable += crossphase_table
            if len(pangle_table) > 0:
                gaintable += pangle_table
        gaintable_bkp = copy.deepcopy(gaintable)
        for g in gaintable_bkp:
            if os.path.basename(g) == "full_selfcal.gcal":
                gaintable.remove(g)
        del gaintable_bkp

        ####################################
        # Filtering any corrupted ms
        #####################################
        filtered_mslist = []  # Filtering in case any ms is corrupted
        for ms in mslist:
            checkcol = check_datacolumn_valid(ms)
            if checkcol:
                filtered_mslist.append(ms)
            else:
                print(f"Issue in : {ms}")
                os.system("rm -rf {ms}")
        mslist = filtered_mslist
        if len(mslist) == 0:
            print("No valid measurement set.")
            print(f"Total time taken: {round(time.time()-start_time,2)}s")
            return 1

        ####################################
        # Applycal jobs
        ####################################
        print(f"Total ms list: {len(mslist)}")
        ms_size_list = [get_column_size(ms) for ms in mslist]
        mem_limit = max(ms_size_list)
        if dask_addr is None:
            dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    len(mslist),
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                )
            )
        else:
            _, _, n_jobs, n_threads, mem_limit, dask_dir = get_dask_client(
                len(mslist),
                dask_dir=workdir,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                min_mem_per_job=mem_limit,
                only_cal=True,
            )
            os.system(f"rm -rf {dask_dir}")
            dask_client = Client(address=dask_addr)
        wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
        tasks = []
        if scaled_bandpass_table != "" and os.path.exists(scaled_bandpass_table):
            bpass_table = scaled_bandpass_table
            if os.path.exists(f"{workdir}/.noattcal"):
                os.system(f"rm -rf {workdir}/.noattcal")
            os.system(f"touch {workdir}/.attcal")
        else:
            print("Bandpass tables with attenuation scaling are not found.")
            if os.path.exists(f"{workdir}/.attcal"):
                os.system(f"rm -rf {workdir}/.attcal")
            os.system(f"touch {workdir}/.noattcal")
            bpass_table = bandpass_table[0]
        for ms in mslist:
            interp = []
            final_gaintable = gaintable + [bpass_table]
            for g in final_gaintable:
                if ".gcal" in g:
                    interp.append("linear")
                elif ".kcal" in g:
                    interp.append("nearest")
                else:
                    interp.append("nearestflag")
            tasks.append(
                delayed(applysol)(
                    ms,
                    gaintable=final_gaintable,
                    overwrite_datacolumn=overwrite_datacolumn,
                    applymode=applymode,
                    interp=interp,
                    do_post_flag=do_post_flag,
                    n_threads=n_threads,
                    parang=parang,
                    memory_limit=mem_limit,
                    force_apply=force_apply,
                )
            )
        futures = dask_client.compute(tasks)
        results = list(dask_client.gather(futures))
        dask_client.close()
        if dask_addr is None:
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
        if np.nansum(results) == 0:
            print("##################")
            print(
                "Applying basic calibration solutions for target scans are done successfully."
            )
            print("Total time taken : ", time.time() - start_time)
            print("##################\n")
            return 0
        else:
            print("##################")
            print(
                "Applying basic calibration solutions for target scans are not done successfully."
            )
            print("Total time taken : ", time.time() - start_time)
            print("##################\n")
            return 1
    except Exception as e:
        traceback.print_exc()
        os.system("rm -rf casa*log")
        print("##################")
        print(
            "Applying basic calibration solutions for target scans are not done successfully."
        )
        print("Total time taken : ", time.time() - start_time)
        print("##################\n")
        return 1


def main(
    mslist,
    workdir,
    caldir,
    use_only_bandpass=False,
    applymode="calflag",
    overwrite_datacolumn=False,
    force_apply=False,
    do_post_flag=False,
    start_remote_log=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    logfile=None,
    jobid=0,
    dask_addr=None,
):
    """
    Apply calibration solutions to a list of measurement sets with optional post-flagging.

    Parameters
    ----------
    mslist : str
        Comma-separated list of measurement set paths to which calibration will be applied.
    workdir : str
        Directory for logs, PID files, and temporary data products.
    caldir : str
        Path to directory containing calibration tables (e.g., bandpass, gain, polarization).
    use_only_bandpass : bool, optional
        If True, applies only the bandpass calibration (no gain or polarization cal). Default is False.
    applymode : str, optional
        CASA calibration application mode (e.g., "calonly", "calflag", "flagonly"). Default is "calflag".
    overwrite_datacolumn : bool, optional
        If True, overwrites the CORRECTED column during calibration. Default is False.
    force_apply : bool, optional
        If True, forces re-application of calibration even if it appears already applied. Default is False.
    do_post_flag : bool, optional
        If True, performs a round of post-calibration flagging. Default is False.
    start_remote_log : bool, optional
        Whether to enable remote logging using job credentials in `workdir`. Default is False.
    cpu_frac : float, optional
        Fraction of CPU resources to allocate per worker. Default is 0.8.
    mem_frac : float, optional
        Fraction of system memory to allocate per worker. Default is 0.8.
    logfile : str or None, optional
        Path to the logfile. If None, logging to file is disabled. Default is None.
    jobid : int, optional
        Identifier for tracking the job and saving PID. Default is 0.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(mslist.split(",")[0])) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "apply_basiccal", logfile, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")
    ############

    try:
        print("\n###################################")
        print("Starting applying solutions...")
        print("###################################\n")

        if caldir == "" or not os.path.exists(caldir):
            print("Provide existing caltable directory.")
            msg = 1
        else:
            mslist = mslist.split(",")
            msg = run_all_applysol(
                mslist,
                workdir,
                caldir,
                use_only_bandpass=use_only_bandpass,
                overwrite_datacolumn=overwrite_datacolumn,
                do_post_flag=do_post_flag,
                applymode=applymode,
                force_apply=force_apply,
                cpu_frac=cpu_frac,
                mem_frac=mem_frac,
                dask_addr=dask_addr,
            )
    except Exception:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        for ms in mslist:
            drop_cache(ms)
        drop_cache(workdir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Apply basic calibration solutions to target scans",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "mslist",
        type=str,
        help="Comma-separated list of measurement sets (required)",
    )
    basic_args.add_argument(
        "--workdir",
        type=str,
        default="",
        required=True,
        help="Working directory for intermediate files",
    )
    basic_args.add_argument(
        "--caldir",
        type=str,
        default="",
        required=True,
        help="Directory containing calibration tables",
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced parameters\n###################"
    )
    adv_args.add_argument(
        "--use_only_bandpass",
        action="store_true",
        help="Use only bandpass calibration solutions",
    )
    adv_args.add_argument(
        "--applymode",
        type=str,
        default="calflag",
        help="Applycal mode (e.g. 'calonly', 'calflag')",
    )
    adv_args.add_argument(
        "--overwrite_datacolumn",
        action="store_true",
        help="Overwrite corrected data column in MS",
    )
    adv_args.add_argument(
        "--force_apply",
        action="store_true",
        help="Force apply calibration even if already applied",
    )
    adv_args.add_argument(
        "--do_post_flag", action="store_true", help="Perform post-calibration flagging"
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac", type=float, default=0.8, help="CPU fraction to use"
    )
    hard_args.add_argument(
        "--mem_frac", type=float, default=0.8, help="Memory fraction to use"
    )
    hard_args.add_argument(
        "--logfile", type=str, default=None, help="Optional path to log file"
    )
    hard_args.add_argument(
        "--jobid", type=str, default="0", help="Job ID for logging and process tracking"
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        args.mslist,
        args.workdir,
        args.caldir,
        use_only_bandpass=args.use_only_bandpass,
        applymode=args.applymode,
        overwrite_datacolumn=args.overwrite_datacolumn,
        force_apply=args.force_apply,
        do_post_flag=args.do_post_flag,
        start_remote_log=args.start_remote_log,
        cpu_frac=float(args.cpu_frac),
        mem_frac=float(args.mem_frac),
        logfile=args.logfile,
        jobid=args.jobid,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\nApplying calibration solutions are done.\n###################\n"
    )
    os._exit(result)



import types
import secrets
import string
import logging
import argparse
import requests
import time
import glob
import sys
import os
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from datetime import datetime as dt
from .basic_utils import *
from .proc_manage_utils import *


##################################
# Logger related functions
##################################
class SmartDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def _get_help_string(self, action):
        # Don't show default for boolean flags
        if isinstance(action, argparse._StoreTrueAction) or isinstance(
            action, argparse._StoreFalseAction
        ):
            return action.help
        return super()._get_help_string(action)


def clean_shutdown(observer):
    if observer:
        observer.stop()
        observer.join(timeout=5)


def generate_password(length=6):
    """
    Generate secure 6-character password with letters, digits, and symbols
    """
    chars = string.ascii_letters + string.digits + "@#$&*"
    return "".join(secrets.choice(chars) for _ in range(length))


def get_remote_logger_link():
    cachedir = get_cachedir()
    username = os.getlogin()
    link_file = os.path.join(cachedir, f"remotelink_{username}.txt")
    for _ in range(5):
        try:
            if os.path.isfile(link_file):
                with open(link_file, "r") as f:
                    lines = [line.strip() for line in f if line.strip()]
                if lines:
                    remote_link = lines[0]
                    if remote_link:
                        try:
                            res = requests.get(remote_link, timeout=2)
                            if res.status_code == 200:
                                return remote_link
                        except Exception:
                            pass
        except Exception:
            pass
        time.sleep(2)
    return ""


def get_emails():
    cachedir = get_cachedir()
    username = os.getlogin()
    email_file = os.path.join(cachedir, f"emails_{username}.txt")
    try:
        with open(email_file, "r") as f:
            lines = [line.strip() for line in f if line.strip()]
    except FileNotFoundError:
        return ""
    if not lines:
        return ""
    else:
        return lines[0]


class StreamToLogger:
    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self._buffer = ""

    def write(self, message):
        # Remove trailing newlines and skip empty messages
        message = message.rstrip()
        if message:
            self.logger.log(self.log_level, message)

    def flush(self):
        pass  # Required for compatibility


class RemoteLogger(logging.Handler):
    """
    Remote logging handler for posting log messages to a web endpoint.
    """

    def __init__(
        self, job_id="default", log_id="run_default", remote_link="", password=""
    ):
        super().__init__()
        self.job_id = job_id
        self.log_id = log_id
        self.password = password
        self.remote_link = remote_link

    def emit(self, record):
        msg = self.format(record)
        try:
            requests.post(
                f"{self.remote_link}/api/log",
                json={
                    "job_id": self.job_id,
                    "log_id": self.log_id,
                    "message": msg,
                    "password": self.password,
                    "first": False,
                },
                timeout=2,
            )
        except Exception as e:
            pass  # Fail silently to avoid interrupting the main app


class LogTailHandler(FileSystemEventHandler):
    """
    Continuous logging
    """

    def __init__(self, logfile, logger):
        self.logfile = logfile
        self.logger = logger
        self._position = os.path.getsize(logfile) if os.path.exists(logfile) else 0

    def on_modified(self, event):
        if event.src_path == self.logfile:
            try:
                with open(self.logfile, "r") as f:
                    f.seek(self._position)
                    lines = f.readlines()
                    self._position = f.tell()
                for line in lines:
                    if line != "" and line != " " and line != "\n":
                        self.logger.info(line.strip())
            except Exception:
                pass


def ping_logger(jobid, remote_jobid, stop_event, remote_link=""):
    """Ping a job-specific keep-alive endpoint periodically until stop_event is set."""
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")
    interval = 10  # 10 min interval
    if remote_link != "":
        url = f"{remote_link}/api/ping/{remote_jobid}"
        while not stop_event.is_set():
            try:
                print(
                    f"[ping_logger] Ping sent for job {remote_jobid} at {dt.now().isoformat()}"
                )
                res = requests.post(url, timeout=2)
            except Exception as e:
                pass
            stop_event.wait(interval)


def create_logger(logname, logfile, get_print=False, verbose=False):
    """
    Create logger.

    Parameters
    ----------
    logname : str
        Name of the log
    logfile : str, optional
        Log file name
    get_print : bool, optional
        Get print output to log
    verbose : bool, optional
        Verbose output or not

    Returns
    -------
    logger
        Python logging object
    str
        Log file name
    """
    if os.path.exists(logfile):
        os.system("rm -rf " + logfile)
    formatter = logging.Formatter("%(message)s")
    logger = logging.getLogger(logname)
    logger.setLevel(logging.DEBUG)
    if verbose:
        console = logging.StreamHandler(sys.stdout)
        console.setFormatter(formatter)
        logger.addHandler(console)
    filehandle = logging.FileHandler(logfile)
    filehandle.setFormatter(formatter)
    logger.addHandler(filehandle)
    logger.propagate = False
    if get_print:
        sys.stdout = StreamToLogger(logger, logging.INFO)
        sys.stderr = StreamToLogger(logger, logging.ERROR)
    logger.info("Log file : " + logfile + "\n")
    return logger, logfile


def get_logid(logfile):
    """
    Get log id for remote logger from logfile name
    """
    name = os.path.basename(logfile)
    logmap = {
        "apply_basiccal.log": "Applying basic calibration solutions",
        "apply_pbcor.log": "Applying primary beam corrections",
        "apply_selfcal.log": "Applying self-calibration solutions",
        "basic_cal.log": "Basic calibration",
        "cor_sidereal_selfcals.log": "Correction of sidereal motion before self-calibration",
        "cor_sidereal_targets.log": "Correction of sidereal motion for target scans",
        "flagging_cal_calibrator.log": "Basic flagging",
        "modeling_calibrator.log": "Simulating visibilities of calibrators",
        "split_targets.log": "Spliting target scans",
        "split_selfcals.log": "Spliting for self-calibration",
        "selfcal_targets.mainlog": "All self-calibrations main log",
        "imaging_targets.mainlog": "All imaging main log",
        "selfcal_targets.log": "All self-calibrations",
        "imaging_targets.log": "All imaging",
        "noise_cal.log": "Flux calibration using noise-diode",
        "partition_cal.log": "Partioning for basic calibration",
        "ds_targets.log": "Making dynamic spectra",
        "main.log":"Main pipeline log",
    }

    if name in logmap:
        return logmap[name]
    elif "selfcals_scan_" in name:
        name = name.rstrip("_selfcal.log")
        scan = name.split("scan_")[-1].split("_spw")[0]
        spw = name.split("spw_")[-1].split("_selfcal")[0]
        return f"Self-calibration for: Scan : {scan}, Spectral window: {spw}"
    elif "imaging_targets_scan_" in name:
        name = name.rstrip(".log")
        scan = name.split("scan_")[-1].split("_spw")[0]
        spw = name.split("spw_")[-1].split("_selfcal")[0]
        return f"Imaging for: Scan : {scan}, Spectral window: {spw}"
    else:
        return name


def init_logger(logname, logfile, jobname="", password=""):
    """
    Initialize a remote logger with watchdog-based tailing.

    Parameters
    ----------
    logname : str
        Logger name.
    logfile : str
        Path to the local logfile to also monitor.
    jobname : str, optional
        Remote logger job ID.
    password : str
        Password used for remote authentication.

    Returns
    -------
    observer 
        Observer object
    """
    timeout = 30
    waited = 0
    while True:
        if os.path.exists(logfile) == False:
            time.sleep(1)
            waited += 1
        elif waited >= timeout:
            return
        else:
            break
    logger = logging.getLogger(logname)
    logger.setLevel(logging.INFO)
    logger.propagate = False
    if logger.hasHandlers():
        logger.handlers.clear()
    formatter = logging.Formatter("%(message)s")
    remote_link = get_remote_logger_link()
    if remote_link != "":
        if jobname:
            job_id = jobname
            log_id = get_logid(logfile)
            remote_handler = RemoteLogger(
                job_id=job_id, log_id=log_id, remote_link=remote_link, password=password
            )
            remote_handler.setFormatter(formatter)
            logger.addHandler(remote_handler)

            try:
                requests.post(
                    f"{remote_link}/api/log",
                    json={
                        "job_id": job_id,
                        "log_id": log_id,
                        "message": "Job starting...",
                        "password": password,
                        "first": True,
                    },
                    timeout=2,
                )
            except Exception:
                pass
        if os.path.exists(logfile):
            event_handler = LogTailHandler(logfile, logger)
            observer = Observer()
            observer.schedule(
                event_handler, path=os.path.dirname(logfile), recursive=False
            )
            observer.start()
            return observer
        else:
            return
    else:
        return


# Exposing only functions
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]



import logging
import dask
import numpy as np
import argparse
import traceback
import time
import sys
import os
from casatasks import casalog
from dask import delayed
from dask.distributed import Client
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)


try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def cor_sidereal_motion(
    mslist, workdir, cpu_frac=0.8, mem_frac=0.8, max_cpu_frac=0.8, max_mem_frac=0.8, dask_addr=None,
):
    """
    Perform sidereal motion correction

    Parameters
    ----------
    mslist : list
        Measurement set list
    workdir : str
        Work directory
    cpu_frac : float, optional
        CPU fraction to use
    mem_frac : float, optional
        Memory fraction to use
    max_cpu_frac : float, optional
        Maximum CPU fraction to use
    max_mem_frac : float, optional
        Maximum memory fraction to use
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    list
        List of sidereal motion corrected measurement sets
    """
    start_time = time.time()
    try:
        container_name = "meerwsclean"
        container_present = check_udocker_container(container_name)
        if not container_present:
            container_name = initialize_wsclean_container(name=container_name)
            if container_name is None:
                print(
                    "Container {container_name} is not initiated. First initiate container and then run."
                )
                return 1, []
        #############################################
        # Memory limit
        #############################################
        mem_limit = correct_solar_sidereal_motion(dry_run=True)
        #############################################
        tasks = []
        for ms in mslist:
            tasks.append(delayed(correct_solar_sidereal_motion)(ms))
        total_chunks = len(tasks)
        if dask_addr is None:
            dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    total_chunks,
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                )
            )
        else:
            _, _, n_jobs, n_threads, mem_limit, dask_dir = (
                get_dask_client(
                    total_chunks,
                    dask_dir=workdir,
                    cpu_frac=cpu_frac,
                    mem_frac=mem_frac,
                    min_mem_per_job=mem_limit,
                    only_cal=True,
                )
            )
            dask_client=Client(address=dask_addr)
            os.system(f"rm -rf {dask_dir}")
        wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
        futures = dask_client.compute(tasks)
        results = list(dask_client.gather(futures))
        dask_client.close()
        if dask_addr is None:
            dask_cluster.close()
            os.system(f"rm -rf {dask_dir}")
        splited_ms_list_phaserotated = []
        for i in range(len(results)):
            msg = results[i]
            ms = mslist[i]
            if msg == 0:
                if os.path.exists(ms + "/.sidereal_cor"):
                    splited_ms_list_phaserotated.append(ms)
        if len(splited_ms_list_phaserotated) == 0:
            print("##################")
            print(
                "Sidereal motion correction is not successful for any measurement set."
            )
            print("Total time taken : ", time.time() - start_time)
            print("##################\n")
            return 1, []
        else:
            print("##################")
            print("Sidereal motion corrections are done successfully.")
            print("Total time taken : ", time.time() - start_time)
            print("##################\n")
            return 0, splited_ms_list_phaserotated
    except Exception as e:
        traceback.print_exc()
        print("##################")
        print("Sidereal motion correction is not successful for any measurement set.")
        print("Total time taken : ", time.time() - start_time)
        print("##################\n")
        return 1, []


def main(
    mslist,
    workdir="",
    cpu_frac=0.8,
    mem_frac=0.8,
    max_cpu_frac=0.8,
    max_mem_frac=0.8,
    logfile=None,
    jobid=0,
    start_remote_log=False,
    dask_addr=None,
):
    """
    Run a parallel processing pipeline for solar sidereal motion correction

    Parameters
    ----------
    mslist : str
        Comma-separated list of paths to measurement sets to be processed.
    workdir : str, optional
        Directory for logs, intermediate files, and other outputs.
        If empty, defaults to the directory of the first MS with `/workdir` appended.
        Default is "".
    cpu_frac : float, optional
        Fraction of total CPU cores to allocate per task. Default is 0.8.
    mem_frac : float, optional
        Fraction of total system memory to allocate per task. Default is 0.8.
    max_cpu_frac : float, optional
        Maximum fraction of total CPU cores to use across all tasks. Default is 0.8.
    max_mem_frac : float, optional
        Maximum fraction of total system memory to use across all tasks. Default is 0.8.
    logfile : str or None, optional
        Path to the log file for capturing logs. If None, logging to file is disabled. Default is None.
    jobid : int, optional
        Unique job identifier used for PID tracking and task differentiation. Default is 0.
    start_remote_log : bool, optional
        Whether to enable remote logging based on credentials stored in the workdir. Default is False.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    mslist = mslist.split(",")
    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(mslist[0])) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and logfile is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            observer = init_logger(
                "do_sidereal_cor", logfile, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")
    ###########

    try:
        if len(mslist) == 0:
            print("Please provide a list of measurement sets.")
            msg = 1
        else:
            msg, final_target_mslist = cor_sidereal_motion(
                mslist,
                workdir,
                cpu_frac=float(cpu_frac),
                mem_frac=float(mem_frac),
                max_cpu_frac=float(max_cpu_frac),
                max_mem_frac=float(max_mem_frac),
                dask_addr=dask_addr,
            )
    except Exception as e:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        for ms in mslist:
            drop_cache(ms)
        drop_cache(workdir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Correct measurement sets for sidereal motion",
        formatter_class=SmartDefaultsHelpFormatter,
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "mslist",
        type=str,
        help="Comma-separated list of measurement sets (required positional argument)",
    )
    basic_args.add_argument("--workdir", type=str, default="", help="Working directory")

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced calibration and imaging parameters\n###################"
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="CPU fraction to use",
        metavar="Float",
    )
    hard_args.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Memory fraction to use",
        metavar="Float",
    )
    hard_args.add_argument(
        "--max_cpu_frac",
        type=float,
        default=0.8,
        help="Maximum CPU fraction to use",
        metavar="Float",
    )
    hard_args.add_argument(
        "--max_mem_frac",
        type=float,
        default=0.8,
        help="Maximum memory fraction to use",
        metavar="Float",
    )
    hard_args.add_argument("--logfile", type=str, default=None, help="Log file")
    hard_args.add_argument("--jobid", type=int, default=0, help="Job ID")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        mslist=args.mslist,
        workdir=args.workdir,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        max_cpu_frac=args.max_cpu_frac,
        max_mem_frac=args.max_mem_frac,
        logfile=args.logfile,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    print(
        "\n###################\\Sidereal motion corrections are done.\n###################\n"
    )
    os._exit(result)



import psutil
import numpy as np
import argparse
import time
import sys
import os
import signal
from casatasks import casalog
from meersolar.utils import get_cachedir, drop_cache

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass


def kill_process_and_children(pid):
    try:
        parent = psutil.Process(pid)
        children = parent.children(recursive=True)
        for child in children:
            try:
                child.kill()
            except Exception:
                pass
        parent.kill()
    except (psutil.NoSuchProcess, ProcessLookupError):
        pass


def force_kill_pids_with_children(pids, max_tries=10, wait_time=0.5):
    """
    Repeatedly try to kill all PIDs (and their children) until none remain.
    """
    for attempt in range(max_tries):
        remaining = []
        for pid in np.atleast_1d(pids):
            try:
                kill_process_and_children(int(pid))
            except Exception:
                remaining.append(pid)

        time.sleep(wait_time)

        remaining = [pid for pid in np.atleast_1d(pids) if psutil.pid_exists(int(pid))]

        if not remaining:
            break
        else:
            pids = remaining


def kill_meerjob():
    """
    Kill MeerSOLAR Job
    """
    parser = argparse.ArgumentParser(description="Kill MeerSOLAR Job")
    parser.add_argument(
        "--jobid", type=str, required=True, help="MeerSOLAR Job ID to kill"
    )
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()
    cachedir = get_cachedir()
    jobfile_name = f"{cachedir}/main_pids_{args.jobid}.txt"
    results = np.loadtxt(jobfile_name, dtype="str", unpack=True)
    main_pid = int(results[1])
    msname = str(results[2])
    workdir = str(results[3])
    outdir = str(results[4])
    pid_file = f"{cachedir}/pids/pids_{args.jobid}.txt"
    try:
        os.kill(int(main_pid), signal.SIGKILL)
    except BaseException:
        pass
    if os.path.exists(pid_file):
        pids = np.loadtxt(pid_file, unpack=True, dtype="int")
        force_kill_pids_with_children(pids)
    os.system(f"rm -rf {workdir}/tmp_meersolar_*")
    drop_cache(msname)
    drop_cache(workdir)
    drop_cache(outdir)
    drop_cache(cachedir)


if __name__ == "__main__":
    kill_meerjob()



import os
import platform
import ctypes
import sys
import numpy as np
import traceback
import argparse
import webbrowser
from collections import deque
from threading import Thread
from PyQt5.QtWidgets import (
    QApplication,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QListWidget,
    QListWidgetItem,
    QPlainTextEdit,
    QSplitter,
    QSizeGrip,
    QGridLayout,
)
from PyQt5.QtCore import Qt, QTimer, pyqtSignal, QObject
from PyQt5.QtGui import QTextCursor
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

LOG_DIR = None
POSIX_FADV_DONTNEED = 4
libc = ctypes.CDLL("libc.so.6")


#####################################
# Resource management
#####################################
def drop_file_cache(filepath, verbose=False):
    """
    Advise the OS to drop the given file from the page cache.
    Safe, per-file, no sudo required.
    """
    if platform.system() != "Linux":
        raise NotImplementedError("drop_file_cache is only supported on Linux")
    try:
        if not os.path.isfile(filepath):
            return
        fd = os.open(filepath, os.O_RDONLY)
        result = libc.posix_fadvise(fd, 0, 0, POSIX_FADV_DONTNEED)
        os.close(fd)
        if verbose:
            if result == 0:
                print(f"[cache drop] Released: {filepath}")
            else:
                print(f"[cache drop] Failed ({result}) for: {filepath}")
    except Exception as e:
        if verbose:
            print(f"[cache drop] Error for {filepath}: {e}")
            traceback.print_exc()


def drop_cache(path, verbose=False):
    """
    Drop file cache for a file or all files under a directory.

    Parameters
    ----------
    path : str
        File or directory path
    """
    if platform.system() != "Linux":
        raise NotImplementedError("drop_file_cache is only supported on Linux")
    if os.path.isfile(path):
        drop_file_cache(path, verbose=verbose)
    elif os.path.isdir(path):
        for root, _, files in os.walk(path):
            for f in files:
                full_path = os.path.join(root, f)
                drop_file_cache(full_path, verbose=verbose)
    else:
        if verbose:
            print(f"[cache drop] Path does not exist or is not valid: {path}")


def get_cachedir():
    homedir = os.environ.get("HOME")
    if homedir is None:
        homedir = os.path.expanduser("~")
    username = os.getlogin()
    cachedir = f"{homedir}/.solarpipe"
    os.makedirs(cachedir, exist_ok=True)
    return cachedir


class SmartDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def _get_help_string(self, action):
        # Don't show default for boolean flags
        if isinstance(action, argparse._StoreTrueAction) or isinstance(
            action, argparse._StoreFalseAction
        ):
            return action.help
        return super()._get_help_string(action)


def get_logid(logfile):
    """
    Get log id for remote logger from logfile name
    """
    name = os.path.basename(logfile)
    logmap = {
        "apply_basiccal.log": "Applying basic calibration solutions",
        "apply_pbcor.log": "Applying primary beam corrections",
        "apply_selfcal.log": "Applying self-calibration solutions",
        "basic_cal.log": "Basic calibration",
        "cor_sidereal_selfcals.log": "Correction of sidereal motion before self-calibration",
        "cor_sidereal_targets.log": "Correction of sidereal motion for target scans",
        "flagging_cal_calibrator.log": "Basic flagging",
        "modeling_calibrator.log": "Simulating visibilities of calibrators",
        "split_targets.log": "Spliting target scans",
        "split_selfcals.log": "Spliting for self-calibration",
        "selfcal_targets.log": "All self-calibrations",
        "imaging_targets.log": "All imaging",
        "noise_cal.log": "Flux calibration using noise-diode",
        "partition_cal.log": "Partioning for basic calibration",
        "ds_targets.log": "Making dynamic spectra",
        "main.log":"Main pipeline logs",
    }

    if name in logmap:
        return logmap[name]
    elif "selfcals_scan_" in name:
        name = name.rstrip("_selfcal.log")
        scan = name.split("scan_")[-1].split("_spw")[0]
        spw = name.split("spw_")[-1].split("_selfcal")[0]
        return f"Self-calibration for: Scan : {scan}, Spectral window: {spw}"
    elif "imaging_targets_scan_" in name:
        name = name.rstrip(".log")
        scan = name.split("scan_")[-1].split("_spw")[0]
        spw = name.split("spw_")[-1].split("_selfcal")[0]
        return f"Imaging for: Scan : {scan}, Spectral window: {spw}"
    else:
        return name


class TailWatcher(FileSystemEventHandler, QObject):
    new_line = pyqtSignal(str)

    def __init__(self, file_path):
        super().__init__()
        self.file_path = file_path
        self._running = True
        self._position = os.path.getsize(file_path) if os.path.exists(file_path) else 0
        self.observer = Observer()

    def start(self):
        self.observer.schedule(
            self, path=os.path.dirname(self.file_path), recursive=False
        )
        self.observer.start()
        # Do not emit initial content to avoid duplication

    def stop(self):
        self._running = False
        self.observer.stop()
        self.observer.join()

    def on_modified(self, event):
        if event.src_path == self.file_path and self._running:
            try:
                with open(self.file_path, "r") as f:
                    f.seek(self._position)
                    new_data = f.read()
                    self._position = f.tell()
                    if new_data:
                        # Remove blank/whitespace-only lines
                        filtered_lines = "\n".join(
                            line for line in new_data.splitlines() if line.strip()
                        )
                        if filtered_lines:
                            self.new_line.emit(filtered_lines + "\n")
            except Exception as e:
                self.new_line.emit(f"\n[watcher error] {e}\n")


class LogViewer(QWidget):
    def __init__(self, max_lines=10000):
        super().__init__()
        self.setWindowTitle("MeerLogger (Live Log)")
        screen = QApplication.primaryScreen().availableGeometry()
        self.setGeometry(
            screen.x() + screen.width() // 10,
            screen.y() + screen.height() // 10,
            int(screen.width() * 0.8),
            int(screen.height() * 0.8),
        )

        self.max_lines = max_lines
        self.buffer = []
        self.tail_watcher = None
        self.current_log_path = None

        self.setup_ui()
        self.refresh_logs()

        self.refresh_timer = QTimer()
        self.refresh_timer.timeout.connect(self.refresh_logs)
        self.refresh_timer.start(2000)

    def calc_list_width(self):
        fm = self.log_list.fontMetrics()
        widths = [
            fm.horizontalAdvance(self.log_list.item(i).text())
            for i in range(self.log_list.count())
        ]
        return max(150, min(int(1.1 * max(widths, default=100)), 500))

    def setup_ui(self):
        outer_layout = QGridLayout(self)
        outer_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(outer_layout)

        inner_layout = QVBoxLayout()
        self.splitter = QSplitter(Qt.Horizontal)
        self.splitter.setChildrenCollapsible(False)

        self.log_list = QListWidget()
        self.log_list.itemClicked.connect(self.load_log_content)

        self.log_view = QPlainTextEdit()
        self.log_view.setReadOnly(True)

        self.splitter.addWidget(self.log_list)
        self.splitter.addWidget(self.log_view)
        inner_layout.addWidget(self.splitter)

        grip_layout = QHBoxLayout()
        grip_layout.addStretch()
        grip_layout.addWidget(QSizeGrip(self))
        inner_layout.addLayout(grip_layout)

        inner_container = QWidget()
        inner_container.setObjectName("InnerContainer")
        inner_container.setLayout(inner_layout)
        inner_container.setStyleSheet(
            """
            QWidget#InnerContainer {
                background-color: #f0f0f0;
                border-bottom-left-radius: 12px;
                border-bottom-right-radius: 12px;
            }
            """
        )

        outer_layout.addWidget(inner_container, 0, 0)

    def refresh_logs(self):
        existing_paths = {
            self.log_list.item(i).data(Qt.UserRole)
            for i in range(self.log_list.count())
        }
        new_items_added = False

        if os.path.isdir(LOG_DIR):
            log_files = [
                fname
                for fname in os.listdir(LOG_DIR)
                if os.path.isfile(os.path.join(LOG_DIR, fname))
                and fname.endswith(".log")
            ]
            log_files.sort(key=lambda f: os.path.getctime(os.path.join(LOG_DIR, f)))

            for fname in log_files:
                full_path = os.path.join(LOG_DIR, fname)
                if full_path not in existing_paths:
                    display_name = get_logid(fname)
                    item = QListWidgetItem(display_name)
                    item.setData(Qt.UserRole, full_path)
                    self.log_list.addItem(item)
                    new_items_added = True

        if new_items_added:
            QTimer.singleShot(
                100,
                lambda: self.splitter.setSizes(
                    [self.calc_list_width(), self.width() - self.calc_list_width()]
                ),
            )

    def load_log_content(self, item):
        new_log_path = item.data(Qt.UserRole)
        if self.tail_watcher:
            self.tail_watcher.stop()
        self.current_log_path = new_log_path
        self.buffer.clear()
        self.log_view.clear()

        try:
            with open(new_log_path, "r") as f:
                full_data = f.read()
                # Split, filter blank lines, and rejoin with original line
                # endings
                lines = [
                    line for line in full_data.splitlines(keepends=True) if line.strip()
                ]
                self.buffer = lines
                self.log_view.setPlainText("".join(lines))
                self.log_view.moveCursor(QTextCursor.End)
        except Exception as e:
            self.buffer = [f"[Error reading file: {e}]\n"]
            self.log_view.setPlainText(self.buffer[0])

        self.tail_watcher = TailWatcher(self.current_log_path)
        self.tail_watcher.new_line.connect(self.append_log_line)
        self.tail_watcher.start()

    def append_log_line(self, text):
        lines = text.splitlines(keepends=True)
        self.buffer.extend(lines)
        if len(self.buffer) > self.max_lines:
            self.buffer = self.buffer[-self.max_lines :]
        self.log_view.setPlainText("".join(self.buffer))
        self.log_view.moveCursor(QTextCursor.End)

    def closeEvent(self, event):
        if self.tail_watcher:
            self.tail_watcher.stop()
        QApplication.quit()


def cli():
    global LOG_DIR
    parser = argparse.ArgumentParser(
        description="MeerSOLAR Logger", formatter_class=SmartDefaultsHelpFormatter
    )
    parser.add_argument("--jobid", type=str, default=None, help="MeerSOLAR Job ID")
    parser.add_argument(
        "--logdir",
        type=str,
        default=None,
        help="Name of log directory",
    )
    parser.add_argument(
        "--no-prefect",
        action="store_false",
        dest="prefect",
        help="Name of log directory",
    )
    args = parser.parse_args()

    if args.prefect:
        cachedir = get_cachedir()
        if os.path.exists(f"{cachedir}/prefect.dashboard"):
            with open(f"{cachedir}/prefect.dashboard","r") as f:
                SERVER_DASHBOARD=f.read()
            webbrowser.open(SERVER_DASHBOARD)
            sys.exit(0)
        else:
            if args.jobid is None and args.logdir is None:
                print("Please provide either job ID or log directory.")
                sys.exit(1)
    else:
        if args.jobid is not None:
            jobfile_name = f"{cachedir}/main_pids_{args.jobid}.txt"
            if not os.path.exists(jobfile_name):
                print(
                    f"Job ID: {args.jobid} is not available. Provide log directory name."
                )
                sys.exit(1)
            else:
                results = np.loadtxt(jobfile_name, dtype="str", unpack=True)
                workdir = str(results[3])
                if not os.path.exists(workdir):
                    print(f"Work directory : {workdir} is not present.")
                    sys.exit(1)
                LOG_DIR = workdir.rstrip("/") + "/logs"
        else:                       
            if not os.path.exists(args.logdir):
                print(
                    f"Log directory: {args.logdir} is not present. Please provide a valid log directory."
                )
                sys.exit(1)
            LOG_DIR = args.logdir

    os.environ["QT_OPENGL"] = "software"
    os.environ["QT_XCB_GL_INTEGRATION"] = "none"
    os.environ["QT_STYLE_OVERRIDE"] = "Fusion"
    os.environ["QT_LOGGING_RULES"] = "qt.qpa.*=false"
    os.makedirs(f"{LOG_DIR}/xdgtmp", exist_ok=True)
    os.chmod(f"{LOG_DIR}/xdgtmp", 0o700)
    os.environ.setdefault("XDG_RUNTIME_DIR", f"{LOG_DIR}/xdgtmp")
    os.environ["XDG_RUNTIME_DIR"] = f"{LOG_DIR}/xdgtmp"
    os.environ["TMPDIR"] = f"{LOG_DIR}/xdgtmp"

    try:
        app = QApplication(sys.argv)
        app.setStyleSheet(
            """
            * {
                font-family: \"Segoe UI\", \"Noto Sans\", \"Sans Serif\";
                font-size: 15px;
            }
            QListWidget, QPlainTextEdit, QPushButton {
                font-size: 15px;
            }
            QPushButton {
                padding: 4px 14px;
            }
        """
        )
        viewer = LogViewer()
        viewer.show()
        sys.exit(app.exec_())
    except Exception as e:
        traceback.print_exc()
    finally:
        if LOG_DIR is not None and os.path.exists(LOG_DIR):
            drop_cache(LOG_DIR)
